/*
 *  Copyright (C) 2003, 2004, 2005, 2006, 2007, 2008, 2009 Apple Inc. All rights reserved.
 *  Copyright (C) 2007 Eric Seidel <eric@webkit.org>
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include "config.h"
#include "MarkedSpace.h"

#include "JSGlobalObject.h"
#include "JSLock.h"
#include "JSObject.h"
#include "ScopeChain.h"

namespace JSC {

class Structure;

#define INIT_EXACT_ALLOCATOR(size) \
    m_exactAllocator##size.init(heap, this, size, false); \
    m_exactDestructorAllocator##size.init(heap, this, size, true);

#define NOT_MULTIPLE_OF_16(size) \
    COMPILE_ASSERT((size & 0xF) != 0, exact_allocator_size_violation);

MarkedSpace::MarkedSpace(Heap* heap)
    : m_heap(heap)
{
    // Make sure we don't have exact allocators for sizes already handled by the "precise" allocators
    FOR_EACH_EXACT_SIZE(NOT_MULTIPLE_OF_16);
    for (size_t cellSize = preciseStep; cellSize <= preciseCutoff; cellSize += preciseStep) {
        allocatorFor(cellSize).init(heap, this, cellSize, false);
        destructorAllocatorFor(cellSize).init(heap, this, cellSize, true);
    }

    for (size_t cellSize = impreciseStep; cellSize <= impreciseCutoff; cellSize += impreciseStep) {
        allocatorFor(cellSize).init(heap, this, cellSize, false);
        destructorAllocatorFor(cellSize).init(heap, this, cellSize, true);
    }

    FOR_EACH_EXACT_SIZE(INIT_EXACT_ALLOCATOR)
}

#define DEFINE_EXACT_ALLOCATOR(size) \
    template<> MarkedAllocator& MarkedSpace::exactAllocator<size>() { return m_exactAllocator##size; } \
    template<> MarkedAllocator& MarkedSpace::exactDestructorAllocator<size>() { return m_exactDestructorAllocator##size; }

FOR_EACH_EXACT_SIZE(DEFINE_EXACT_ALLOCATOR)

#define RESET_EXACT_ALLOCATOR(size) \
    m_exactAllocator##size.reset(); \
    m_exactDestructorAllocator##size.reset();

void MarkedSpace::resetAllocators()
{
    for (size_t cellSize = preciseStep; cellSize <= preciseCutoff; cellSize += preciseStep) {
        allocatorFor(cellSize).reset();
        destructorAllocatorFor(cellSize).reset();
    }

    for (size_t cellSize = impreciseStep; cellSize <= impreciseCutoff; cellSize += impreciseStep) {
        allocatorFor(cellSize).reset();
        destructorAllocatorFor(cellSize).reset();
    }

    FOR_EACH_EXACT_SIZE(RESET_EXACT_ALLOCATOR)
}

#define ZAP_EXACT_ALLOCATOR(size) \
    m_exactAllocator##size.zapFreeList(); \
    m_exactDestructorAllocator##size.zapFreeList();

void MarkedSpace::canonicalizeCellLivenessData()
{
    for (size_t cellSize = preciseStep; cellSize <= preciseCutoff; cellSize += preciseStep) {
        allocatorFor(cellSize).zapFreeList();
        destructorAllocatorFor(cellSize).zapFreeList();
    }

    for (size_t cellSize = impreciseStep; cellSize <= impreciseCutoff; cellSize += impreciseStep) {
        allocatorFor(cellSize).zapFreeList();
        destructorAllocatorFor(cellSize).zapFreeList();
    }

    FOR_EACH_EXACT_SIZE(ZAP_EXACT_ALLOCATOR)
}

#define RETURN_TRUE_IF_EXACT_ALLOCATOR_PAGED_OUT(size) \
    if (m_exactAllocator##size.isPagedOut(deadline) || m_exactDestructorAllocator##size.isPagedOut(deadline)) \
        return true;

bool MarkedSpace::isPagedOut(double deadline)
{
    for (size_t cellSize = preciseStep; cellSize <= preciseCutoff; cellSize += preciseStep) {
        if (allocatorFor(cellSize).isPagedOut(deadline) || destructorAllocatorFor(cellSize).isPagedOut(deadline))
            return true;
    }

    for (size_t cellSize = impreciseStep; cellSize <= impreciseCutoff; cellSize += impreciseStep) {
        if (allocatorFor(cellSize).isPagedOut(deadline) || destructorAllocatorFor(cellSize).isPagedOut(deadline))
            return true;
    }

    FOR_EACH_EXACT_SIZE(RETURN_TRUE_IF_EXACT_ALLOCATOR_PAGED_OUT)

    return false;
}

void MarkedSpace::freeBlocks(MarkedBlock* head)
{
    MarkedBlock* next;
    for (MarkedBlock* block = head; block; block = next) {
        next = static_cast<MarkedBlock*>(block->next());
        
        m_blocks.remove(block);
        block->sweep();

        m_heap->blockAllocator().deallocate(block);
    }
}

class TakeIfUnmarked {
public:
    typedef MarkedBlock* ReturnType;
    
    TakeIfUnmarked(MarkedSpace*);
    void operator()(MarkedBlock*);
    ReturnType returnValue();
    
private:
    MarkedSpace* m_markedSpace;
    DoublyLinkedList<MarkedBlock> m_empties;
};

inline TakeIfUnmarked::TakeIfUnmarked(MarkedSpace* newSpace)
    : m_markedSpace(newSpace)
{
}

#define EXACT_ALLOCATOR_REMOVE_BLOCK(size) \
    if (block->cellSize() == size) { \
        if (block->cellsNeedDestruction()) \
            m_markedSpace->m_exactDestructorAllocator##size.removeBlock(block); \
        else \
            m_markedSpace->m_exactAllocator##size.removeBlock(block); \
    } else \

inline void TakeIfUnmarked::operator()(MarkedBlock* block)
{
    if (!block->markCountIsZero())
        return;
    
#ifdef GIBBON_SCRIPT_JSC_EXACT_ALLOCATORS
    FOR_EACH_EXACT_SIZE(EXACT_ALLOCATOR_REMOVE_BLOCK)   // else
        m_markedSpace->allocatorForBlock(block).removeBlock(block);
#else
    m_markedSpace->allocatorForBlock(block).removeBlock(block);
#endif
    m_empties.append(block);
}

inline TakeIfUnmarked::ReturnType TakeIfUnmarked::returnValue()
{
    return m_empties.head();
}

void MarkedSpace::shrink()
{
    // We record a temporary list of empties to avoid modifying m_blocks while iterating it.
    TakeIfUnmarked takeIfUnmarked(this);
    freeBlocks(forEachBlock(takeIfUnmarked));
}

#if ENABLE(GGC)
class GatherDirtyCells {
    WTF_MAKE_NONCOPYABLE(GatherDirtyCells);
public:
    typedef void* ReturnType;
    
    explicit GatherDirtyCells(MarkedBlock::DirtyCellVector*);
    void operator()(MarkedBlock*);
    ReturnType returnValue() { return 0; }
    
private:
    MarkedBlock::DirtyCellVector* m_dirtyCells;
};

inline GatherDirtyCells::GatherDirtyCells(MarkedBlock::DirtyCellVector* dirtyCells)
    : m_dirtyCells(dirtyCells)
{
}

inline void GatherDirtyCells::operator()(MarkedBlock* block)
{
    block->gatherDirtyCells(*m_dirtyCells);
}

void MarkedSpace::gatherDirtyCells(MarkedBlock::DirtyCellVector& dirtyCells)
{
    GatherDirtyCells gatherDirtyCells(&dirtyCells);
    forEachBlock(gatherDirtyCells);
}
#endif

} // namespace JSC
